from __future__ import annotations

from typing import TYPE_CHECKING, cast

from gpgi._data_types import Dataset, Geometry, Grid, ParticleSet

if TYPE_CHECKING:
    from typing import Any, Literal

    from gpgi._typing import FieldMap, FloatT, GridDict, ParticleSetDict


def load(
    *,
    geometry: Literal["cartesian", "polar", "cylindrical", "spherical", "equatorial"],
    grid: GridDict[FloatT],
    particles: ParticleSetDict[FloatT] | None = None,
    metadata: dict[str, Any] | None = None,
) -> Dataset[FloatT]:
    r"""
    Load a Dataset.

    Parameters
    ----------
    geometry: Literal["cartesian", "polar", "cylindrical", "spherical", "equatorial"]
        This flag is used for validation of axis names, order and domain limits.

    grid: dict[str, FieldMap]
        A dictionary representing the grid coordinates as 1D arrays of cell left edges,
        and on-grid fields as ND arrays (fields are assumed to be defined on cell
        centers)

    particles: dict[str, FieldMap] (optional)
        A dictionary representing particle coordinates and associated fields as 1D
        arrays

    metadata: dict[str, Any] (optional)
        A dictionary representing arbitrary additional data, that will be attached to
        the returned Dataset as an attribute (namely, ds.metadata). This special
        attribute is accessible from boundary condition methods as the argument of the
        same name.
    """
    try:
        # when Python 3.11 is dropped, this try-except can be simplified as
        # if geometry not in Geometry: raise
        _geometry = Geometry(geometry)
    except ValueError:
        # this exception is raised directly, not added to a stack,
        # in order to avoid "possibly unbound" warnings from static analysis
        raise ValueError(
            f"unknown geometry {geometry!r}, "
            f"expected any of {tuple(_.value for _ in Geometry)}"
        ) from None

    exceptions: list[Exception] = []

    if "cell_edges" not in grid:
        exceptions.append(
            ValueError("grid dictionary missing required key 'cell_edges'")
        )
    if particles is not None and "coordinates" not in particles:
        exceptions.append(
            ValueError("particles dictionary missing required key 'coordinates'")
        )
    if len(exceptions) == 1:
        raise exceptions[0]
    elif exceptions:
        raise ExceptionGroup("Invalid inputs were received", exceptions)

    _grid = Grid(
        geometry=_geometry,
        cell_edges=cast("FieldMap[FloatT]", grid["cell_edges"]),
        fields=grid.get("fields"),
    )

    _particles: ParticleSet[FloatT] | None = None
    if particles is not None:
        _particles = ParticleSet(
            geometry=_geometry,
            coordinates=cast("FieldMap[FloatT]", particles["coordinates"]),
            fields=particles.get("fields"),
        )

    return Dataset(
        geometry=_geometry, grid=_grid, particles=_particles, metadata=metadata
    )



import sys
from importlib.metadata import version
from importlib.util import find_spec

import pytest

import gpgi

HAVE_PYTEST_MPL = find_spec("pytest_mpl") is not None


def pytest_configure(config):
    if HAVE_PYTEST_MPL:
        import matplotlib as mpl

        mpl.use("Agg")
    else:  # pragma: no cover
        config.addinivalue_line(
            "markers", "mpl_image_compare: skip (missing requirement: pytest_mpl)"
        )


def pytest_runtest_setup(item):
    if HAVE_PYTEST_MPL:
        return
    if any(item.iter_markers(name="mpl_image_compare")):  # pragma: no cover
        pytest.skip("missing requirement: pytest_mpl")


def pytest_report_header(config, start_path):
    is_gil_enabled = sys.version_info < (3, 13) or sys._is_gil_enabled()

    return [
        f"{is_gil_enabled = }",
        f"NumPy: {version('numpy')}",
        f"{gpgi._IS_PY_LIB = }",
        f"gpgi._lib loads from {find_spec('gpgi._lib').origin}",
    ]



from typing import Generic, NotRequired, TypedDict, TypeVar

import numpy as np
from numpy.typing import NDArray

FloatT = TypeVar("FloatT", np.float32, np.float64)
HCIArray = NDArray[np.uint16]


Name = str
FieldMap = dict[str, NDArray[FloatT]]


class CartesianCoordinates(TypedDict, Generic[FloatT]):
    x: NDArray[FloatT]
    y: NotRequired[NDArray[FloatT]]
    z: NotRequired[NDArray[FloatT]]


class CylindricalCoordinates(TypedDict, Generic[FloatT]):
    radius: NDArray[FloatT]
    azimuth: NotRequired[NDArray[FloatT]]
    z: NotRequired[NDArray[FloatT]]


class PolarCoordinates(TypedDict, Generic[FloatT]):
    radius: NDArray[FloatT]
    z: NotRequired[NDArray[FloatT]]
    azimuth: NotRequired[NDArray[FloatT]]


class SphericalCoordinates(TypedDict, Generic[FloatT]):
    colatitude: NDArray[FloatT]
    radius: NotRequired[NDArray[FloatT]]
    azimuth: NotRequired[NDArray[FloatT]]


class EquatorialCoordinates(TypedDict, Generic[FloatT]):
    radius: NDArray[FloatT]
    latitude: NotRequired[NDArray[FloatT]]
    azimuth: NotRequired[NDArray[FloatT]]


CoordMap = (
    CartesianCoordinates[FloatT]
    | CylindricalCoordinates[FloatT]
    | PolarCoordinates[FloatT]
    | SphericalCoordinates[FloatT]
    | EquatorialCoordinates[FloatT]
)


class GridDict(TypedDict, Generic[FloatT]):
    cell_edges: CoordMap[FloatT]
    fields: NotRequired[FieldMap[FloatT]]


class ParticleSetDict(TypedDict, Generic[FloatT]):
    coordinates: CoordMap[FloatT]
    fields: NotRequired[FieldMap[FloatT]]



# cython: freethreading_compatible = True
# note that this macro is defined only to prevent the interpreter from
# re-enabling the GIL when this extension is imported from, but the functions
# themselves are absolutely not thread-safe since they work by mutating objects
# that can be accessed from another thread (namely, the `out` argument).
# This design choice was made to remove the overhead of memory allocation in
# performance measurements. It is fine to keep this interface as long as the out
# argument isn't directly exposed in public API.

cimport cython
cimport numpy as np
from libc.math cimport floor, fmin

cdef fused real:
    np.float64_t
    np.float32_t


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _index_particles(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] dx,
    np.ndarray[np.uint16_t, ndim=2] out,
):
    r"""
    compute the host cell index (*out*)
    The result is a mapping from particle index to containing cell index.
    Note that cell edges are to be treated as ordained from left to right
    and are padded with ghost layers (one for each side and direction).
    Particles on the other hand are not assumed to be sorted, but they *are*
    assumed to be all contained in the active domain.

    dx is a 3 element array that contains the constant step used in each direction,
    if any, or -1 as a filler value otherwise.

    In directions where a constant step is used, the result is immediate (O(1))
    Otherwise we use a bisection search (O(log(N)))
    """
    cdef Py_ssize_t ipart, particle_count
    cdef np.uint16_t iL, iR, idx

    # reduce the chance of floating point arithmetic errors
    # by forcing positions be read as double precision
    # (instead of reading as the input type)
    cdef np.float64_t x

    cdef int ndim
    if cell_edges_x3.shape[0] > 1:
        ndim = 3
    elif cell_edges_x2.shape[0] > 1:
        ndim = 2
    else:
        ndim = 1

    particle_count = particles_x1.shape[0]

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            if dx[0] > 0:
                out[ipart, 0] = int((x - cell_edges_x1[0]) // dx[0])
            else:
                iL = 1
                iR = cell_edges_x1.shape[0] - 2
                idx = (iL + iR) // 2
                while idx != iL:
                    if cell_edges_x1[idx] > x:
                        iR = idx
                    else:
                        iL = idx
                    idx = (iL + iR) // 2
                out[ipart, 0] = idx

            if ndim < 2:
                continue

            x = particles_x2[ipart]
            if dx[1] > 0:
                out[ipart, 1] = int((x - cell_edges_x2[0]) // dx[1])
            else:
                iL = 1
                iR = cell_edges_x2.shape[0] - 2
                idx = (iL + iR) // 2
                while idx != iL:
                    if cell_edges_x2[idx] > x:
                        iR = idx
                    else:
                        iL = idx
                    idx = (iL + iR) // 2
                out[ipart, 1] = idx

            if ndim < 3:
                continue

            x = particles_x3[ipart]
            if dx[2] > 0:
                out[ipart, 2] = int((x - cell_edges_x3[0]) // dx[2])
            else:
                iL = 1
                iR = cell_edges_x3.shape[0] - 2
                idx = (iL + iR) // 2
                while idx != iL:
                    if cell_edges_x3[idx] > x:
                        iR = idx
                    else:
                        iL = idx
                    idx = (iL + iR) // 2
                out[ipart, 2] = idx


@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_ngp_1D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=1] out,
):
    cdef Py_ssize_t ipart, particle_count
    particle_count = hci.shape[0]

    cdef int i

    cdef np.uint16_t[:, :] hci_v = hci
    cdef real[:] field_v = field
    cdef real[:] out_v = out

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            i = hci_v[ipart][0]
            if no_weight:
                out_v[i] += field_v[ipart]
            else:
                out_v[i] += field_v[ipart] * wfield_v[ipart]


@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_ngp_2D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=2] out,
):
    cdef Py_ssize_t ipart, particle_count
    particle_count = hci.shape[0]

    cdef int i, j

    cdef np.uint16_t[:, :] hci_v = hci
    cdef real[:] field_v = field
    cdef real[:, :] out_v = out

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            i = hci_v[ipart][0]
            j = hci_v[ipart][1]
            if no_weight:
                out_v[i][j] += field_v[ipart]
            else:
                out_v[i][j] += field_v[ipart] * wfield_v[ipart]


@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_ngp_3D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=3] out,
):
    cdef Py_ssize_t ipart, particle_count
    particle_count = hci.shape[0]

    cdef int i, j, k

    cdef np.uint16_t[:, :] hci_v = hci
    cdef real[:] field_v = field
    cdef real[:, :, :] out_v = out

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            i = hci_v[ipart][0]
            j = hci_v[ipart][1]
            k = hci_v[ipart][2]
            if no_weight:
                out_v[i][j][k] += field_v[ipart]
            else:
                out_v[i][j][k] += field_v[ipart] * wfield_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_cic_1D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=1] out,
):

    cdef Py_ssize_t ipart, particle_count, i, oci
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, ci_start

    cdef real[:] field_v = field
    cdef real[:] out_v = out

    # geometric weight arrays
    cdef real[2] w

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            ci = hci[ipart, 0]
            d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
            ci_start = ci + <np.uint16_t>fmin(0, floor(2*d)-1)
            w[0] = 0.5 - d + fmin(1, floor(2*d))
            w[1] = 1 - w[0]

            for i in range(2):
                oci = ci_start + i
                if no_weight:
                    out_v[oci] += w[i] * field_v[ipart]
                else:
                    out_v[oci] += w[i] * field_v[ipart] * wfield_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_cic_2D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=2] out,
):

    cdef Py_ssize_t ipart, particle_count, i, j, oci, ocj
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, cj, ci_start, cj_start

    cdef real[:] field_v = field
    cdef real[:, :] out_v = out

    # geometric weight arrays
    cdef real[2] w1, w2
    cdef real[2][2] w

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            ci = hci[ipart, 0]
            d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
            ci_start = ci + <np.uint16_t>fmin(0, floor(2*d)-1)
            w1[0] = 0.5 - d + fmin(1, floor(2*d))
            w1[1] = 1 - w1[0]

            x = particles_x2[ipart]
            cj = hci[ipart, 1]
            d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
            cj_start = cj + <np.uint16_t>fmin(0, floor(2*d)-1)
            w2[0] = 0.5 - d + fmin(1, floor(2*d))
            w2[1] = 1 - w2[0]

            for i in range(2):
                for j in range(2):
                    w[i][j] = w1[i] * w2[j]

            for i in range(2):
                oci = ci_start + i
                for j in range(2):
                    ocj = cj_start + j
                    if no_weight:
                        out_v[oci, ocj] += w[i][j] * field_v[ipart]
                    else:
                        out_v[oci, ocj] += w[i][j] * field_v[ipart] * wfield_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_cic_3D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=3] out,
):

    cdef Py_ssize_t ipart, particle_count, i, j, k, oci, ocj, ock
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, cj, ck, ci_start, cj_start, ck_start

    cdef real[:] field_v = field
    cdef real[:, :, :] out_v = out

    # geometric weight arrays
    cdef real[2] w1, w2, w3
    cdef real[2][2][2] w

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            ci = hci[ipart, 0]
            d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
            ci_start = ci + <np.uint16_t>fmin(0, floor(2*d)-1)
            w1[0] = 0.5 - d + fmin(1, floor(2*d))
            w1[1] = 1 - w1[0]

            x = particles_x2[ipart]
            cj = hci[ipart, 1]
            d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
            cj_start = cj + <np.uint16_t>fmin(0, floor(2*d)-1)
            w2[0] = 0.5 - d + fmin(1, floor(2*d))
            w2[1] = 1 - w2[0]

            x = particles_x3[ipart]
            ck = hci[ipart, 2]
            d = (x - cell_edges_x3[ck]) / (cell_edges_x3[ck + 1] - cell_edges_x3[ck])
            ck_start = ck + <np.uint16_t>fmin(0, floor(2*d)-1)
            w3[0] = 0.5 - d + fmin(1, floor(2*d))
            w3[1] = 1 - w3[0]

            for i in range(2):
                for j in range(2):
                    for k in range(2):
                        w[i][j][k] = w1[i] * w2[j] * w3[k]

            for i in range(2):
                oci = ci_start + i
                for j in range(2):
                    ocj = cj_start + j
                    for k in range(2):
                        ock = ck_start + k
                        if no_weight:
                            out_v[oci, ocj, ock] += w[i][j][k] * field_v[ipart]
                        else:
                            out_v[oci, ocj, ock] += w[i][j][k] * field_v[ipart] * wfield_v[ipart]  # noqa E501


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_tsc_1D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=1] out,
):
    cdef Py_ssize_t ipart, particle_count, i, oci
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci

    cdef real[:] field_v = field
    cdef real[:] out_v = out

    # geometric weight arrays
    cdef real[3] w

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            ci = hci[ipart, 0]
            d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
            w[0] = 0.5 * (1 - d) ** 2
            w[1] = 0.75 - (d - 0.5) ** 2
            w[2] = 0.5 * d**2

            for i in range(3):
                oci = ci - 1 + i
                if no_weight:
                    out_v[oci] += w[i] * field_v[ipart]
                else:
                    out_v[oci] += w[i] * field_v[ipart] * wfield_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_tsc_2D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=2] out,
):
    cdef Py_ssize_t ipart, particle_count, i, j, oci, ocj
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, cj

    cdef real[:] field_v = field
    cdef real[:, :] out_v = out

    # geometric weight arrays
    cdef real[3] w1, w2
    cdef real[3][3] w

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            ci = hci[ipart, 0]
            d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
            w1[0] = 0.5 * (1 - d) ** 2
            w1[1] = 0.75 - (d - 0.5) ** 2
            w1[2] = 0.5 * d**2

            x = particles_x2[ipart]
            cj = hci[ipart, 1]
            d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
            w2[0] = 0.5 * (1 - d) ** 2
            w2[1] = 0.75 - (d - 0.5) ** 2
            w2[2] = 0.5 * d**2

            for i in range(3):
                for j in range(3):
                    w[i][j] = w1[i] * w2[j]

            for i in range(3):
                oci = ci - 1 + i
                for j in range(3):
                    ocj = cj - 1 + j
                    if no_weight:
                        out_v[oci, ocj] += w[i][j] * field_v[ipart]
                    else:
                        out_v[oci, ocj] += w[i][j] * field_v[ipart] * wfield_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_tsc_3D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[real, ndim=1] weight_field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=3] out,
):
    cdef Py_ssize_t ipart, particle_count, i, j, k, oci, ocj, ock
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, cj, ck

    cdef real[:] field_v = field
    cdef real[:, :, :] out_v = out

    # geometric weight arrays
    cdef real[3] w1, w2, w3
    cdef real[3][3][3] w

    cdef int lw = weight_field.shape[0]
    cdef real[:] wfield_v = weight_field
    cdef bint no_weight = (lw == 0)

    with nogil:
        for ipart in range(particle_count):
            x = particles_x1[ipart]
            ci = hci[ipart, 0]
            d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
            w1[0] = 0.5 * (1 - d) ** 2
            w1[1] = 0.75 - (d - 0.5) ** 2
            w1[2] = 0.5 * d**2

            x = particles_x2[ipart]
            cj = hci[ipart, 1]
            d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
            w2[0] = 0.5 * (1 - d) ** 2
            w2[1] = 0.75 - (d - 0.5) ** 2
            w2[2] = 0.5 * d**2

            x = particles_x3[ipart]
            ck = hci[ipart, 2]
            d = (x - cell_edges_x3[ck]) / (cell_edges_x3[ck + 1] - cell_edges_x3[ck])
            w3[0] = 0.5 * (1 - d) ** 2
            w3[1] = 0.75 - (d - 0.5) ** 2
            w3[2] = 0.5 * d**2

            for i in range(3):
                for j in range(3):
                    for k in range(3):
                        w[i][j][k] = w1[i] * w2[j] * w3[k]

            for i in range(3):
                oci = ci - 1 + i
                for j in range(3):
                    ocj = cj - 1 + j
                    for k in range(3):
                        ock = ck - 1 + k
                        if no_weight:
                            out_v[oci, ocj, ock] += w[i][j][k] * field_v[ipart]
                        else:
                            out_v[oci, ocj, ock] += w[i][j][k] * field_v[ipart] * wfield_v[ipart]  # noqa E501



import math
from dataclasses import dataclass
from enum import StrEnum, auto
from itertools import chain
from typing import Any, Protocol, TypeVar, assert_never, final

import numpy as np
from numpy.typing import NDArray

from gpgi._typing import FieldMap, FloatT, Name


class Geometry(StrEnum):
    CARTESIAN = auto()
    POLAR = auto()
    CYLINDRICAL = auto()
    SPHERICAL = auto()
    EQUATORIAL = auto()


_AXES_LIMITS: dict[Name, tuple[float, float]] = {
    "x": (-float("inf"), float("inf")),
    "y": (-float("inf"), float("inf")),
    "z": (-float("inf"), float("inf")),
    "radius": (0, float("inf")),
    "azimuth": (0, 2 * np.pi),
    "colatitude": (0, np.pi),
    "latitude": (-np.pi / 2, np.pi / 2),
}


class SpatialData(Protocol):
    geometry: Geometry
    axes: tuple[Name, ...]
    coordinates: FieldMap
    fields: FieldMap


T_contra = TypeVar("T_contra", bound=SpatialData, contravariant=True)


class Validator(Protocol[T_contra]):
    @classmethod
    def collect_exceptions(cls, data: T_contra) -> list[Exception]: ...


@final
class GeometryValidator:
    @classmethod
    def collect_exceptions(cls, data: SpatialData) -> list[Exception]:
        retv: list[Exception] = []
        match data.geometry:
            case Geometry.CARTESIAN:
                axes3D = ("x", "y", "z")
            case Geometry.POLAR:
                axes3D = ("radius", "azimuth", "z")
            case Geometry.CYLINDRICAL:
                axes3D = ("radius", "z", "azimuth")
            case Geometry.SPHERICAL:
                axes3D = ("radius", "colatitude", "azimuth")
            case Geometry.EQUATORIAL:
                axes3D = ("radius", "azimuth", "latitude")
            case _ as unreachable:  # pragma: no cover
                assert_never(unreachable)

        def position(i: int) -> str:
            return ["first", "second", "third"][i]

        for i, (expected, actual) in enumerate(zip(axes3D, data.axes, strict=False)):
            if actual != expected:
                retv.append(
                    ValueError(
                        f"Invalid {position(i)} axis name {actual!r}, "
                        f"with geometry {data.geometry.name.lower()!r}. "
                        f"Got {actual!r}, expected {expected!r}"
                    )
                )
        return retv


@final
class BasicCoordinatesValidator:
    @classmethod
    def collect_exceptions(cls, data: SpatialData) -> list[Exception]:
        retv: list[Exception] = []
        for axis in data.axes:
            coord = data.coordinates[axis]
            if len(coord) == 0:
                continue

            if coord.dtype.kind != "f":
                retv.append(
                    ValueError(
                        f"Invalid data type {coord.dtype} (expected a float dtype)"
                    )
                )
                continue
            dt = coord.dtype.type

            xmin, xmax = (dt(_) for _ in _AXES_LIMITS[axis])
            if (cmin := dt(np.min(coord))) < xmin or not math.isfinite(cmin):
                if math.isfinite(xmin):
                    hint = f"minimal value allowed is {xmin}"
                else:
                    assert xmin == -float("inf")
                    hint = "value must be finite"
                retv.append(
                    ValueError(
                        f"Invalid coordinate data for axis {axis!r} {cmin} ({hint})"
                    )
                )
            if (cmax := dt(np.max(coord))) > xmax or not math.isfinite(cmax):
                if math.isfinite(xmax):
                    hint = f"maximal value allowed is {xmax}"
                else:
                    assert xmax == float("inf")
                    hint = "value must be finite"
                retv.append(
                    ValueError(
                        f"Invalid coordinate data for axis {axis!r} {cmax} ({hint})"
                    )
                )
        return retv


@final
class DTypeConsistencyValidator:
    @classmethod
    def collect_exceptions(cls, data: SpatialData) -> list[Exception]:
        retv: list[Exception] = []
        dts = {
            name: arr.dtype
            for name, arr in chain(
                data.coordinates.items(),
                data.fields.items(),
            )
        }
        unique_dts = sorted(set(dts.values()))
        if len(unique_dts) > 1:
            retv.append(TypeError(f"Received mixed data types ({unique_dts}):\n{dts}"))
        return retv


@final
@dataclass(frozen=True, slots=True)
class NamedArray:
    name: Name
    data: NDArray


@final
class FieldMapsValidatorHelper:
    @classmethod
    def collect_exceptions(
        cls,
        *fmaps: FieldMap[FloatT] | None,
        require_shape_equality: bool = False,
        require_sorted: bool = False,
        required_attrs: dict[str, Any] | None = None,
    ) -> list[Exception]:
        ref_arr: NamedArray | None = None
        retv: list[Exception] = []
        for name, data in chain.from_iterable(
            fm.items() for fm in fmaps if fm is not None
        ):
            if require_shape_equality:
                try:
                    ref_arr = cls._check_shape_equality(name, data, ref_arr)
                except ValueError as exc:
                    retv.append(exc)
                    continue
            if require_sorted:
                if (result := cls._check_sorted_state(name, data)) is not None:
                    retv.append(result)
            if required_attrs:
                if (
                    result := cls._check_required_attributes(name, data, required_attrs)
                ) is not None:
                    retv.append(result)
        return retv

    @staticmethod
    def _check_shape_equality(
        name: str,
        data: NDArray[FloatT],
        ref_arr: NamedArray | None,
    ) -> NamedArray:
        if ref_arr is not None and data.shape != ref_arr.data.shape:
            raise ValueError(
                f"Fields {name!r} and {ref_arr.name!r} "
                f"have mismatching shapes {data.shape} and {ref_arr.data.shape}"
            )
        return ref_arr or NamedArray(name, data)

    @staticmethod
    def _check_sorted_state(name: str, data: NDArray[FloatT]) -> ValueError | None:
        a = data[0]
        for i, b in enumerate(data[1:], start=1):
            if a > b:
                return ValueError(
                    f"Field {name!r} is not properly sorted by ascending order. "
                    f"Got {a} (index {i - 1}) > {b} (index {i})"
                )
            a = b

        return None

    @staticmethod
    def _check_required_attributes(
        name: str,
        data: NDArray[FloatT],
        required_attrs: dict[str, Any],
    ) -> ValueError | None:
        for attr, expected in required_attrs.items():
            if (actual := getattr(data, attr)) != expected:
                return ValueError(
                    f"Field {name!r} has incorrect {attr} {actual} "
                    f"(expected {expected})"
                )
        return None



from collections.abc import Callable
from threading import Lock
from typing import Any, Literal, cast

from numpy.typing import NDArray

from gpgi._typing import FloatT

BoundaryRecipeT = Callable[
    [
        NDArray[FloatT],
        NDArray[FloatT],
        NDArray[FloatT],
        NDArray[FloatT],
        NDArray[FloatT],
        NDArray[FloatT],
        NDArray[FloatT],
        NDArray[FloatT],
        Literal["left", "right"],
        dict[str, Any],
    ],
    NDArray[FloatT],
]


class BoundaryRegistry:
    def __init__(self) -> None:
        self._registry: dict[str, BoundaryRecipeT] = {}
        self._lock = Lock()
        for key, recipe in _base_registry.items():
            self.register(key, recipe, skip_validation=True)

    @staticmethod
    def _validate_recipe(recipe: BoundaryRecipeT) -> None:
        import inspect

        sig = inspect.signature(recipe)
        params = [
            name for name, p in sig.parameters.items() if p.default == inspect._empty
        ]
        if params != [
            "same_side_active_layer",
            "same_side_ghost_layer",
            "opposite_side_active_layer",
            "opposite_side_ghost_layer",
            "weight_same_side_active_layer",
            "weight_same_side_ghost_layer",
            "weight_opposite_side_active_layer",
            "weight_opposite_side_ghost_layer",
            "side",
            "metadata",
        ]:
            raise ValueError(
                "Invalid boundary recipe. Expected a function with exactly 6 parameters, "
                "named 'same_side_active_layer', 'same_side_ghost_layer', "
                "'opposite_side_active_layer', 'opposite_side_ghost_layer', "
                "'weight_same_side_active_layer', 'weight_same_side_ghost_layer', "
                "'weight_opposite_side_active_layer', 'weight_opposite_side_ghost_layer', "
                "'side', and 'metadata'"
            )

    def register(
        self,
        key: str,
        recipe: BoundaryRecipeT,
        *,
        skip_validation: bool = False,
        allow_unsafe_override: bool = False,
    ) -> None:
        """
        Register a new boundary function.

        Parameters
        ----------
        key: str
            A unique identifier (ideally a meaningful name) to associate with
            the function.

        recipe: Callable
            A function matching the signature (order and names of arguments) of
            gpgi's builtin boundary recipes.

        skip_validation: bool, optional, keyword-only (default: False)
            If set to True, signature validation is skipped.
            This is meant to allow bypassing hypothetical bugs in the validation
            routine.

        allow_unsafe_override: bool, optional, keyword-only (default: False)
            If set to True, registering a new function under an existing key
            will not raise an exception. Note however that doing so is not
            thread-safe.

        Raises
        ------
        ValueError:
        - if skip_validation==False and the signature of the recipe doesn't meet
          the requirements.
        - if allow_unsafe_override==False and a new function is being registered
          under an already used key. Registering the same exact function under
          multiple times either under the same key or another, unused key, is
          always safe so it does not raise.
        """
        with self._lock:
            if key in self._registry:
                if recipe is self._registry[key]:
                    return
                elif not allow_unsafe_override:
                    raise ValueError(
                        f"Another function is already registered with {key=!r}. "
                        "If you meant to override the existing function, "
                        "consider setting allow_unsafe_override=True"
                    )

            if not skip_validation:
                self._validate_recipe(recipe)

            self._registry[key] = recipe

    def __getitem__(self, key: str) -> BoundaryRecipeT:
        return self._registry[key]

    def __contains__(self, key: str) -> bool:
        return key in self._registry


# basic recipes
def open_boundary(
    same_side_active_layer: NDArray[FloatT],
    same_side_ghost_layer: NDArray[FloatT],
    opposite_side_active_layer: NDArray[FloatT],
    opposite_side_ghost_layer: NDArray[FloatT],
    weight_same_side_active_layer: NDArray[FloatT],
    weight_same_side_ghost_layer: NDArray[FloatT],
    weight_opposite_side_active_layer: NDArray[FloatT],
    weight_opposite_side_ghost_layer: NDArray[FloatT],
    side: Literal["left", "right"],
    metadata: dict[str, Any],
) -> NDArray[FloatT]:
    # return the active layer unchanged
    return same_side_active_layer


def wall_boundary(
    same_side_active_layer: NDArray[FloatT],
    same_side_ghost_layer: NDArray[FloatT],
    opposite_side_active_layer: NDArray[FloatT],
    opposite_side_ghost_layer: NDArray[FloatT],
    weight_same_side_active_layer: NDArray[FloatT],
    weight_same_side_ghost_layer: NDArray[FloatT],
    weight_opposite_side_active_layer: NDArray[FloatT],
    weight_opposite_side_ghost_layer: NDArray[FloatT],
    side: Literal["left", "right"],
    metadata: dict[str, Any],
) -> NDArray[FloatT]:
    return cast("NDArray[FloatT]", same_side_active_layer + same_side_ghost_layer)


def antisymmetric_boundary(
    same_side_active_layer: NDArray[FloatT],
    same_side_ghost_layer: NDArray[FloatT],
    opposite_side_active_layer: NDArray[FloatT],
    opposite_side_ghost_layer: NDArray[FloatT],
    weight_same_side_active_layer: NDArray[FloatT],
    weight_same_side_ghost_layer: NDArray[FloatT],
    weight_opposite_side_active_layer: NDArray[FloatT],
    weight_opposite_side_ghost_layer: NDArray[FloatT],
    side: Literal["left", "right"],
    metadata: dict[str, Any],
) -> NDArray[FloatT]:
    return cast("NDArray[FloatT]", same_side_active_layer - same_side_ghost_layer)


def periodic_boundary(
    same_side_active_layer: NDArray[FloatT],
    same_side_ghost_layer: NDArray[FloatT],
    opposite_side_active_layer: NDArray[FloatT],
    opposite_side_ghost_layer: NDArray[FloatT],
    weight_same_side_active_layer: NDArray[FloatT],
    weight_same_side_ghost_layer: NDArray[FloatT],
    weight_opposite_side_active_layer: NDArray[FloatT],
    weight_opposite_side_ghost_layer: NDArray[FloatT],
    side: Literal["left", "right"],
    metadata: dict[str, Any],
) -> NDArray[FloatT]:
    return cast("NDArray[FloatT]", same_side_active_layer + opposite_side_ghost_layer)


_base_registry: dict[str, BoundaryRecipeT] = {
    "open": open_boundary,
    "wall": wall_boundary,
    "antisymmetric ": antisymmetric_boundary,
    "periodic": periodic_boundary,
}



# exercise some thread concurrency
# https://py-free-threading.github.io/debugging/


import threading
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import numpy.testing as npt
import pytest

import gpgi
from gpgi._boundaries import BoundaryRegistry

prng = np.random.default_rng()


N_THREADS = 10


def random_dataset():
    return gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(-0.1, 1.1, 13),
                "y": np.linspace(-0.1, 1.1, 13),
            },
        },
        particles={
            "coordinates": {
                "x": prng.random(10_000),
                "y": prng.random(10_000),
            },
            "fields": {
                "mass": prng.random(10_000),
            },
        },
    )


class TestSetupHostCellIndex:
    def check(self, results):
        # Check results: verify that all threads see the same array
        assert len(set(results)) == 1
        # check that we indeed collected ids, not None
        assert isinstance(results[0], int)

    def test_concurrent_threading(self):
        # Defines a thread barrier that will be spawned before parallel execution
        # this increases the probability of concurrent access clashes.
        barrier = threading.Barrier(N_THREADS)

        # This object will be shared by all the threads.
        ds = random_dataset()

        results = []

        def closure():
            # Ensure that all threads reach this point before concurrent execution.
            barrier.wait()
            hci = ds._setup_host_cell_index()
            assert ds.host_cell_index is hci
            results.append(id(hci))

        # Spawn n threads that call _setup_host_cell_index concurrently.
        workers = []
        for _ in range(0, N_THREADS):
            workers.append(threading.Thread(target=closure))

        for worker in workers:
            worker.start()

        for worker in workers:
            worker.join()

        self.check(results)

    def test_concurrent_pool(self):
        # Defines a thread barrier that will be spawned before parallel execution
        # this increases the probability of concurrent access clashes.
        barrier = threading.Barrier(N_THREADS)

        # This object will be shared by all the threads.
        ds = random_dataset()

        def closure():
            # Ensure that all threads reach this point before concurrent execution.
            barrier.wait()
            hci = ds._setup_host_cell_index()
            assert ds.host_cell_index is hci
            return id(hci)

        with ThreadPoolExecutor(max_workers=N_THREADS) as executor:
            futures = [executor.submit(closure) for _ in range(N_THREADS)]

        results = [f.result() for f in futures]
        self.check(results)


@pytest.mark.parametrize("lock", ["per-instance", None, threading.Lock()])
class TestDeposit:
    def check(self, results):
        ref = results[0]
        for res in results[1:]:
            npt.assert_array_equal(res, ref)
        assert len({id(res) for res in results}) == len(results)

    def test_concurrent_threading(self, lock):
        # Defines a thread barrier that will be spawned before parallel execution
        # this increases the probability of concurrent access clashes.
        barrier = threading.Barrier(N_THREADS)

        # This object will be shared by all the threads.
        ds = random_dataset()

        results = []

        def closure():
            # Ensure that all threads reach this point before concurrent execution.
            barrier.wait()
            dep = ds.deposit("mass", method="nearest_grid_point", lock=lock)
            results.append(dep)

        # Spawn n threads that call _setup_host_cell_index concurrently.
        workers = []
        for _ in range(0, N_THREADS):
            workers.append(threading.Thread(target=closure))

        for worker in workers:
            worker.start()

        for worker in workers:
            worker.join()

        self.check(results)

    def test_concurrent_pool(self, lock):
        # Defines a thread barrier that will be spawned before parallel execution
        # this increases the probability of concurrent access clashes.
        barrier = threading.Barrier(N_THREADS)

        # This object will be shared by all the threads.
        ds = random_dataset()

        def closure():
            # Ensure that all threads reach this point before concurrent execution.
            barrier.wait()
            return ds.deposit("mass", method="nearest_grid_point", lock=lock)

        with ThreadPoolExecutor(max_workers=N_THREADS) as executor:
            futures = [executor.submit(closure) for _ in range(N_THREADS)]

        results = [f.result() for f in futures]
        self.check(results)


class TestBoundaryRegistry:
    def check(self, results):
        # only one thread can succeed registration, all others should raise.
        expected_msg = (
            "Another function is already registered with key='test'. "
            "If you meant to override the existing function, "
            "consider setting allow_unsafe_override=True"
        )
        assert len(results) == N_THREADS - 1
        assert results.count(expected_msg) == N_THREADS - 1

    def test_concurrent_threading(self):
        # Defines a thread barrier that will be spawned before parallel execution
        # this increases the probability of concurrent access clashes.
        barrier = threading.Barrier(N_THREADS)

        # This object will be shared by all the threads.
        registry = BoundaryRegistry()

        results = []

        def closure():
            def test_recipe(
                same_side_active_layer,
                same_side_ghost_layer,
                opposite_side_active_layer,
                opposite_side_ghost_layer,
                weight_same_side_active_layer,
                weight_same_side_ghost_layer,
                weight_opposite_side_active_layer,
                weight_opposite_side_ghost_layer,
                side,
                metadata,
            ): ...

            assert "test" not in registry

            # Ensure that all threads reach this point before concurrent execution.
            barrier.wait()
            try:
                registry.register("test", test_recipe)
            except ValueError as exc:
                msg, *_ = exc.args
                results.append(msg)
            assert "test" in registry

        # Spawn n threads that call _setup_host_cell_index concurrently.
        workers = []
        for _ in range(0, N_THREADS):
            workers.append(threading.Thread(target=closure))

        for worker in workers:
            worker.start()

        for worker in workers:
            worker.join()

        self.check(results)

    def test_concurrent_pool(self):
        # Defines a thread barrier that will be spawned before parallel execution
        # this increases the probability of concurrent access clashes.
        barrier = threading.Barrier(N_THREADS)

        # This object will be shared by all the threads.
        registry = BoundaryRegistry()

        def closure():
            def test_recipe(
                same_side_active_layer,
                same_side_ghost_layer,
                opposite_side_active_layer,
                opposite_side_ghost_layer,
                weight_same_side_active_layer,
                weight_same_side_ghost_layer,
                weight_opposite_side_active_layer,
                weight_opposite_side_ghost_layer,
                side,
                metadata,
            ): ...

            assert "test" not in registry

            # Ensure that all threads reach this point before concurrent execution.
            barrier.wait()
            registry.register("test", test_recipe)

        with ThreadPoolExecutor(max_workers=N_THREADS) as executor:
            futures = [executor.submit(closure) for _ in range(N_THREADS)]

        assert "test" in registry
        exceptions = [f.exception() for f in futures]
        results = [exc.args[0] for exc in exceptions if exc is not None]
        self.check(results)



r"""Define the core data structures of the library: Grid, ParticleSet, and Dataset."""

from __future__ import annotations

import sys
import warnings
from contextlib import AbstractContextManager, nullcontext
from copy import deepcopy
from enum import Enum, auto
from functools import cached_property, partial, reduce
from textwrap import indent
from threading import Lock
from time import monotonic_ns
from typing import TYPE_CHECKING, Generic, Literal, cast, final

import numpy as np

from gpgi._boundaries import BoundaryRegistry
from gpgi._lib import (
    _deposit_cic_1D,
    _deposit_cic_2D,
    _deposit_cic_3D,
    _deposit_ngp_1D,
    _deposit_ngp_2D,
    _deposit_ngp_3D,
    _deposit_tsc_1D,
    _deposit_tsc_2D,
    _deposit_tsc_3D,
    _index_particles,
)
from gpgi._spatial_data import (
    BasicCoordinatesValidator,
    DTypeConsistencyValidator,
    FieldMapsValidatorHelper,
    Geometry,
    GeometryValidator,
    Validator,
)
from gpgi._typing import FieldMap, FloatT, Name

if sys.version_info >= (3, 13):
    LockType = Lock
else:
    from _thread import LockType

if TYPE_CHECKING:
    from typing import Any, Self

    from numpy.typing import NDArray

    from gpgi._typing import HCIArray
    from gpgi.typing import DepositionMethodT, DepositionMethodWithMetadataT


BoundarySpec = tuple[tuple[str, str, str], ...]


class DepositionMethod(Enum):
    NEAREST_GRID_POINT = auto()
    CLOUD_IN_CELL = auto()
    TRIANGULAR_SHAPED_CLOUD = auto()


_deposition_method_names: dict[str, DepositionMethod] = {
    **{m.name.lower(): m for m in DepositionMethod},
    **{"".join([w[0] for w in m.name.split("_")]).lower(): m for m in DepositionMethod},
}


_BUILTIN_METHODS: dict[DepositionMethod, list[DepositionMethodT]] = {
    DepositionMethod.NEAREST_GRID_POINT: [
        _deposit_ngp_1D,
        _deposit_ngp_2D,
        _deposit_ngp_3D,
    ],
    DepositionMethod.CLOUD_IN_CELL: [
        _deposit_cic_1D,
        _deposit_cic_2D,
        _deposit_cic_3D,
    ],
    DepositionMethod.TRIANGULAR_SHAPED_CLOUD: [
        _deposit_tsc_1D,
        _deposit_tsc_2D,
        _deposit_tsc_3D,
    ],
}


# the following need to be defined in the same module as Grid and ParticleSet
@final
class GridCoordinatesValidator:
    @classmethod
    def collect_exceptions(cls, data: Grid[FloatT]) -> list[Exception]:
        return FieldMapsValidatorHelper.collect_exceptions(
            data.coordinates,
            require_sorted=True,
            required_attrs={"ndim": 1},
        )


@final
class GridFieldsValidator:
    @classmethod
    def collect_exceptions(cls, data: Grid[FloatT]) -> list[Exception]:
        return FieldMapsValidatorHelper.collect_exceptions(
            data.fields,
            required_attrs={
                "size": data.size,
                "ndim": data.ndim,
                "shape": data.shape,
            },
        )


@final
class Grid(Generic[FloatT]):
    def __init__(
        self,
        *,
        geometry: Geometry,
        cell_edges: FieldMap[FloatT],
        fields: FieldMap[FloatT] | None = None,
    ) -> None:
        r"""
        Define a Grid from cell left-edges and data fields.

        Parameters
        ----------
        geometry (keyword-only): gpgi.Geometry

        cell_edges (keyword-only): gpgi.typing.FieldMap
            Left cell edges in each direction as 1D arrays, including the right edge
            of the rightmost cell.

        fields (keyword-only, optional): gpgi.typing.FieldMap
        """
        self.geometry = geometry
        self.coordinates: FieldMap[FloatT] = cell_edges

        if fields is None:
            fields = {}
        self.fields: FieldMap[FloatT] = fields

        self.axes = tuple(self.coordinates.keys())
        self._validate()
        self.dtype: np.dtype[FloatT] = self.coordinates[self.axes[0]].dtype

        self._dx: NDArray[FloatT] = np.full(
            (3,), -1, dtype=self.coordinates[self.axes[0]].dtype
        )
        for i, ax in enumerate(self.axes):
            if self.size == 1 or np.diff(self.coordinates[ax]).std() < 1e-16:
                # got a constant step in this direction, store it
                self._dx[i] = self.coordinates[ax][1] - self.coordinates[ax][0]

    _validators: list[type[Validator[Grid[FloatT]]]] = [
        GeometryValidator,
        BasicCoordinatesValidator,
        GridCoordinatesValidator,
        GridFieldsValidator,
        DTypeConsistencyValidator,
    ]

    def _validate(self) -> None:
        exceptions: list[Exception] = []
        for validator in self.__class__._validators:
            exceptions.extend(validator.collect_exceptions(self))
        if len(exceptions) == 1:
            raise exceptions[0]
        elif exceptions:
            raise ExceptionGroup("input grid data is invalid", exceptions)

    def __repr__(self) -> str:
        """Implement repr(Grid(...))."""
        return (
            f"{self.__class__.__name__}(\n"
            f"    geometry={str(self.geometry)!r},\n"
            f"    cell_edges={self.coordinates},\n"
            f"    fields={self.fields},\n"
            ")"
        )

    @property
    def cell_edges(self) -> FieldMap[FloatT]:
        r"""An alias for self.coordinates."""
        return self.coordinates

    @cached_property
    def cell_centers(self) -> FieldMap[FloatT]:
        r"""The positions of cell centers in each direction."""
        return {ax: 0.5 * (arr[1:] + arr[:-1]) for ax, arr in self.coordinates.items()}  # type: ignore [misc]

    @cached_property
    def cell_widths(self) -> FieldMap[FloatT]:
        r"""The width of cells, expressed as the difference between consecutive left edges."""
        return {ax: np.diff(arr) for ax, arr in self.coordinates.items()}

    @cached_property
    def shape(self) -> tuple[int, ...]:
        r"""The shape of the grid, as a tuple (nx1, (nx2, (nx3)))."""
        return tuple(len(_) - 1 for _ in self.cell_edges.values())

    @cached_property
    def _padded_shape(self) -> tuple[int, ...]:
        return tuple(len(_) + 1 for _ in self.cell_edges.values())

    @property
    def size(self) -> int:
        r"""The total number of cells in the grid."""
        return reduce(int.__mul__, self.shape)

    @property
    def ndim(self) -> int:
        r"""The number of spatial dimensions that coordinates are defined in."""
        return len(self.axes)

    @property
    def cell_volumes(self) -> NDArray[FloatT]:
        r"""
        The generalized ND-volume of grid cells.

        3D: (nx, ny, nz) array representing volumes
        2D: (nx, ny) array representing surface
        1D: (nx,) array representing widths (redundant with cell_widths).
        """
        widths = list(self.cell_widths.values())
        if self.geometry is Geometry.CARTESIAN:
            raw = np.prod(np.meshgrid(*widths), axis=0)
            return np.swapaxes(raw, 0, 1)
        else:
            raise NotImplementedError(
                f"cell_volumes property is not implemented for {self.geometry} geometry"
            )


@final
class ParticleSetCoordinatesValidator:
    @classmethod
    def collect_exceptions(cls, data: ParticleSet[FloatT]) -> list[Exception]:
        return FieldMapsValidatorHelper.collect_exceptions(
            data.coordinates,
            require_shape_equality=True,
            required_attrs={"ndim": 1},
        )


@final
class ParticleSet(Generic[FloatT]):
    def __init__(
        self,
        *,
        geometry: Geometry,
        coordinates: FieldMap[FloatT],
        fields: FieldMap[FloatT] | None = None,
    ) -> None:
        r"""
        Define a ParticleSet from point positions and data fields.

        Parameters
        ----------
        geometry (keyword-only): gpgi.Geometry

        coordinates (keyword-only): gpgi.typing.FieldMap
            Particle positions in each direction as 1D arrays.

        fields (keyword-only, optional): gpgi.typing.FieldMap
        """
        self.geometry = geometry
        self.coordinates: FieldMap[FloatT] = coordinates

        if fields is None:
            fields = {}
        self.fields: FieldMap[FloatT] = fields

        self.axes = tuple(self.coordinates.keys())
        self._validate()
        self.dtype: np.dtype[FloatT] = self.coordinates[self.axes[0]].dtype

    _validators: list[type[Validator[ParticleSet[FloatT]]]] = [
        GeometryValidator,
        BasicCoordinatesValidator,
        ParticleSetCoordinatesValidator,
        DTypeConsistencyValidator,
    ]

    def _validate(self) -> None:
        exceptions: list[Exception] = []
        for validator in self.__class__._validators:
            exceptions.extend(validator.collect_exceptions(self))
        if len(exceptions) == 1:
            raise exceptions[0]
        elif exceptions:
            raise ExceptionGroup("input particle data is invalid", exceptions)

    def __repr__(self) -> str:
        """Implement repr(ParticleSet(...))."""
        return (
            f"{self.__class__.__name__}(\n"
            f"    geometry={str(self.geometry)!r},\n"
            f"    coordinates={self.coordinates},\n"
            f"    fields={self.fields},\n"
            ")"
        )

    @property
    def count(self) -> int:
        r"""The total number of particles in the set."""
        return len(next(iter(self.coordinates.values())))

    @property
    def ndim(self) -> int:
        r"""The number of spatial dimensions that coordinates are defined in."""
        return len(self.axes)


@final
class Dataset(Generic[FloatT]):
    def __init__(
        self,
        *,
        geometry: Geometry = Geometry.CARTESIAN,
        grid: Grid[FloatT],
        particles: ParticleSet[FloatT] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        r"""
        Compose a Dataset from a Grid and a ParticleSet.

        Parameters
        ----------
        geometry (keyword-only): gpgi.Geometry
            An enum member that represents the geometry.

        grid (keyword-only): gpgi.Grid

        particles (keyword-only, optional): gpgi.ParticleSet

        metadata (keyword-only, optional): dict[str, Any]
            A dictionary representing arbitrary additional data, that will be attached
            to the returned Dataset as an attribute (namely, ds.metadata). This special
            attribute is accessible from boundary condition methods as the argument of
            the same name.

            .. versionadded: 0.4.0
        """
        self.geometry = geometry

        if particles is None:
            particles = ParticleSet(
                geometry=grid.geometry,
                coordinates={ax: np.array([], dtype=grid.dtype) for ax in grid.axes},
            )

        self.grid: Grid[FloatT] = grid
        self.particles: ParticleSet[FloatT] = particles

        self.boundary_recipes = BoundaryRegistry()
        self.axes = self.grid.axes
        self.metadata = deepcopy(metadata) if metadata is not None else {}

        self._hci: HCIArray | None = None
        self._hci_lock = Lock()
        self._deposit_lock = Lock()

        self._validate()

    def __repr__(self) -> Name:
        """Implement repr(Dataset(...))."""
        return (
            f"{self.__class__.__name__}(\n"
            f"    geometry={str(self.geometry)!r},\n"
            f"    grid={indent(str(self.grid), ' ' * 4).lstrip()},\n"
            f"    particles={indent(str(self.particles), ' ' * 4).lstrip()},\n"
            f"    metadata={self.metadata},\n"
            ")"
        )

    def _validate(self) -> None:
        if self.particles.count == 0:
            return
        for ax, edges in self.grid.cell_edges.items():
            domain_left = float(edges[0])
            domain_right = float(edges[-1])
            for x in self.particles.coordinates[ax]:
                if x < domain_left:
                    raise ValueError(f"Got particle at {ax}={x} < {domain_left=}")
                if x > domain_right:
                    raise ValueError(f"Got particle at {ax}={x} > {domain_right=}")

        if self.particles.dtype != self.grid.dtype:
            raise TypeError(
                f"Got mixed data types:\n"
                f"- from grid data: {self.grid.dtype}\n"
                f"- from particles: {self.particles.dtype}\n"
            )

    def _get_padded_cell_edges(
        self,
    ) -> tuple[NDArray[FloatT], NDArray[FloatT], NDArray[FloatT]]:
        edges = iter(self.grid.cell_edges.values())

        def pad(a: NDArray[FloatT]) -> NDArray[FloatT]:
            dx = a[1] - a[0]
            return np.concatenate([[a[0] - dx], a, [a[-1] + dx]])

        x1 = next(edges)
        cell_edges_x1 = pad(x1)
        DTYPE = cell_edges_x1.dtype

        cell_edges_x2: np.ndarray[tuple[int, ...], np.dtype[FloatT]] = np.empty(
            0, DTYPE
        )
        cell_edges_x3: np.ndarray[tuple[int, ...], np.dtype[FloatT]] = np.empty(
            0, DTYPE
        )
        if self.grid.ndim >= 2:
            cell_edges_x2 = pad(next(edges))
        if self.grid.ndim == 3:
            cell_edges_x3 = pad(next(edges))

        return cell_edges_x1, cell_edges_x2, cell_edges_x3

    def _get_3D_particle_coordinates(
        self,
    ) -> tuple[NDArray[FloatT], NDArray[FloatT], NDArray[FloatT]]:
        particle_coords = iter(self.particles.coordinates.values())
        particles_x1 = next(particle_coords)
        DTYPE = particles_x1.dtype

        particles_x2 = particles_x3 = np.empty(0, DTYPE)
        # mypy is disabled here because at the time of writing, there doesn't
        # seem to be reliable support for expressing dimensionality of ndarrays
        # through type hints, which would in principle allow to resolve the
        # problem it is flagging.
        if self.grid.ndim >= 2:
            particles_x2 = next(particle_coords)  # type: ignore [arg-type]
        if self.grid.ndim == 3:
            particles_x3 = next(particle_coords)  # type: ignore [arg-type]

        return particles_x1, particles_x2, particles_x3

    def _compute_host_cell_index(self, verbose: bool = False) -> HCIArray:
        hci = np.empty((self.particles.count, self.grid.ndim), dtype="uint16")

        tstart = monotonic_ns()
        _index_particles(
            *self._get_padded_cell_edges(),
            *self._get_3D_particle_coordinates(),
            dx=self.grid._dx,
            out=hci,
        )
        for idim in range(hci.shape[1]):
            # There are at least two edge cases where we need clipping to correct raw indices:
            # - particles that live exactly on the domain right edge will be indexed out of bound
            # - single precision positions can lead to overshoots by simple effect of floating point arithmetics
            #   (double precision isn't safe either, it's just safer (by a lot))
            hci[:, idim].clip(1, self.grid.shape[idim], out=hci[:, idim])

        tstop = monotonic_ns()
        if verbose:
            print(
                f"Indexed {self.particles.count:.4g} particles in {(tstop - tstart) / 1e9:.2f} s"
            )
        return hci

    def _setup_host_cell_index(self, verbose: bool = False) -> HCIArray:
        """Pre-compute internal host cell index array, used for depostition.

        This method is thread-safe.
        The result is returned for testing/typechecking convenience.
        """
        # the first thread to acquire the lock does the computation, all other
        # threads can skip it and reuse the result
        with self._hci_lock:
            if self._hci is None:
                self._hci = self._compute_host_cell_index(verbose)
        return self._hci

    @property
    def host_cell_index(self) -> HCIArray:
        r"""
        The ND index of the host cell for each particle.

        It has shape (particles.count, grid.ndim).
        Indices are 0-based and ghost layers are included.
        """
        if self._hci is None:
            # rebinding for typechecking convenience only
            self._hci = self._setup_host_cell_index()
        return self._hci

    @property
    def _default_sort_axes(self) -> tuple[int, ...]:
        return tuple(range(self.grid.ndim - 1, -1, -1))

    def _validate_sort_axes(self, axes: tuple[int, ...]) -> None:
        if len(axes) != self.grid.ndim:
            raise ValueError(f"Expected exactly {self.grid.ndim} axes, got {len(axes)}")
        if any(not isinstance(axis, int) for axis in axes):
            raise ValueError(
                f"Expected all axes to be integers, got {axes!r} "
                f"with types ({', '.join([type(axis).__name__ for axis in axes])})"
            )
        if any(axis > self.grid.ndim - 1 for axis in axes):
            raise ValueError(f"Expected all axes to be <{self.grid.ndim}, got {axes!r}")

    def _get_sort_key(self, axes: tuple[int, ...]) -> NDArray[np.uint16]:
        self._validate_sort_axes(axes)

        hci = self.host_cell_index
        if self.grid.ndim == 1:
            ind = np.lexsort((hci[:, axes[0]],))
        elif self.grid.ndim == 2:
            ind = np.lexsort((hci[:, axes[0]], hci[:, axes[1]]))
        else:
            ind = np.lexsort((hci[:, axes[0]], hci[:, axes[1]], hci[:, axes[2]]))
        return np.array(ind, dtype="uint16")

    def is_sorted(self, *, axes: tuple[int, ...] | None = None) -> bool:
        r"""
        Return True if and only if particles are already sorted.

        .. versionadded: 0.14.0

        Parameters
        ----------
        axes (keyword-only, optional): tuple[int, ...]
            specify in which order axes should be used for sorting.
        """
        sort_key = self._get_sort_key(axes or self._default_sort_axes)
        hci = self.host_cell_index
        return bool(np.all(hci == hci[sort_key]))

    def sorted(self, *, axes: tuple[int, ...] | None = None) -> Self:
        r"""
        Return a copy of this dataset with particles sorted by host cell index.

        .. versionadded: 0.14.0

        Parameters
        ----------
        axes (keyword-only, optional): tuple[int, ...]
            specify in which order axes should be used for sorting.
        """
        sort_key = self._get_sort_key(axes or self._default_sort_axes)

        return type(self)(
            geometry=self.geometry,
            grid=self.grid,
            particles=ParticleSet(
                geometry=self.geometry,
                coordinates={
                    name: arr[sort_key]
                    for name, arr in deepcopy(self.particles.coordinates).items()
                },
                fields={
                    name: arr[sort_key]
                    for name, arr in deepcopy(self.particles.fields).items()
                },
            ),
        )

    def deposit(
        self,
        particle_field_key: Name,
        /,
        *,
        method: Literal[
            "ngp",
            "nearest_grid_point",
            "cic",
            "cloud_in_cell",
            "tsc",
            "triangular_shaped_cloud",
        ]
        | DepositionMethodT
        | DepositionMethodWithMetadataT,
        boundaries: dict[Name, tuple[Name, Name]] | None = None,
        verbose: bool = False,
        return_ghost_padded_array: bool = False,
        weight_field: Name | None = None,
        weight_field_boundaries: dict[Name, tuple[Name, Name]] | None = None,
        lock: Literal["per-instance"] | None | LockType = "per-instance",
    ) -> NDArray[FloatT]:
        r"""
        Perform particle deposition and return the result as a grid field.

        Parameters
        ----------
        particle_field_key (positional only): str
            label of the particle field to deposit

        method (keyword only): 'ngp', 'cic' or 'tsc', or function
            full names ('nearest_grid_point', 'cloud_in_cell', and
            'triangular_shaped_cloud') are also valid

            .. versionchanged:: 0.12.0
                Added support for user-defined functions.

        verbose (keyword only, optional): bool (default False)
            if True, print execution time for hot loops (indexing and deposition)

        return_ghost_padded_array (keyword only, optional): bool (default False)
            if True, return the complete deposition array, including one extra
            cell layer per direction and per side. This option is meant as a
            debugging tool for methods that leak some particle data outside the
            active domain (cic and tsc).

        weight_field (keyword only, optional): str
            label of another field to use as weights. Let u be the field to
            deposit and w be the weight field. Let u' and w' be their equivalent
            on-grid descriptions. u' is obtained as

            w'(x) = Σ w(i) c(i,x)
            u'(x) = (1/w'(x)) Σ u(i) w(i) c(i,x)

            where x is the spatial position, i is a particle index, and w(i,x)
            are geometric coefficients associated with the deposition method.

            .. versionadded: 0.7.0

        boundaries and weight_field_boundaries (keyword only, optional): dict
            Maps from axis names (str) to boundary recipe keys (str, str)
            representing left/right boundaries. By default all axes will use
            'open' boundaries on both sides. Specifying boundaries for all axes
            is not mandated, but note that recipes are applied in the order of
            specified axes (any unspecified axes will be treated last).

            weight_field_boundaries is required if weight field is used in
            combinations with boundaries.

            Boundary recipes are applied the weight field (if any) first.

            .. versionadded: 0.5.0

        lock (keyword only, optional): 'per-instance' (default), None, or threading.Lock
            Fine tune performance for multi-threaded applications: define a
            locking strategy around the deposition hotloop.
            - 'per-instance': allow multiple Dataset instances to run deposition
                concurrently, but forbid concurrent accesses to any specific
                instance
            - None: no locking is applied. Within some restricted conditions
                (e.g. depositing a couple fields concurrently in a sorted dataset),
                this may improve walltime performance, but it is also expected to
                degrade it in a more general case as it encourages cache-misses
            - an arbitrary threading.Lock instance may be supplied to implement
                a custom strategy

            .. versionadded:: 2.0.0
        """
        if callable(method):
            from inspect import signature

            sig = signature(method)
            func: DepositionMethodT
            if "metadata" in sig.parameters:
                method = cast("DepositionMethodWithMetadataT", method)
                func = partial(method, metadata=self.metadata)
            else:
                method = cast("DepositionMethodT", method)
                func = method
        else:
            if method not in _deposition_method_names:
                raise ValueError(
                    f"Unknown deposition method {method!r}, "
                    f"expected any of {tuple(_deposition_method_names.keys())}"
                )

            if (mkey := _deposition_method_names[method]) not in _BUILTIN_METHODS:
                raise NotImplementedError(f"method {method} is not implemented yet")

            func = _BUILTIN_METHODS[mkey][self.grid.ndim - 1]

        if self.particles.count == 0:
            raise TypeError("Cannot deposit particle fields on a particle-less dataset")
        if not self.particles.fields:
            raise TypeError("There are no particle fields")
        if particle_field_key not in self.particles.fields:
            raise ValueError(f"Unknown particle field {particle_field_key!r}")

        if boundaries is None:
            boundaries = {}
        if weight_field_boundaries is None:
            if boundaries and weight_field is not None:
                raise TypeError(
                    "weight_field_boundaries keyword argument is "
                    "required with weight_field and boundaries"
                )
            weight_field_boundaries = {}
        elif weight_field is None:
            warnings.warn(
                "weight_field_boundaries will not be used "
                "as no weight_field was specified",
                stacklevel=2,
            )

        self._sanitize_boundaries(boundaries)
        self._sanitize_boundaries(weight_field_boundaries)

        lock_ctx: AbstractContextManager
        match lock:
            case "per-instance":
                lock_ctx = self._deposit_lock
            case None:
                lock_ctx = nullcontext()
            case LockType():
                lock_ctx = lock
            case _:
                raise ValueError(
                    f"Received {lock=!r}. Expected either 'per-instance', "
                    "None, or an instance of threading.Lock"
                )

        field = self.particles.fields[particle_field_key]
        padded_ret_array = np.zeros(self.grid._padded_shape, dtype=field.dtype)
        if weight_field is not None:
            wfield = np.array(self.particles.fields[weight_field])
            wfield_dep = np.zeros(self.grid._padded_shape, dtype=field.dtype)
        else:
            wfield = np.array((), dtype=field.dtype)
            wfield_dep = np.array((1,), dtype=field.dtype)

        # rebinding for typechecking convenience only
        self._hci = self._setup_host_cell_index(verbose)

        tstart = monotonic_ns()
        with lock_ctx:
            if weight_field is not None:
                func(
                    *self._get_padded_cell_edges(),
                    *self._get_3D_particle_coordinates(),
                    wfield,
                    np.array((), dtype=field.dtype),
                    self._hci,
                    wfield_dep,
                )

            func(
                *self._get_padded_cell_edges(),
                *self._get_3D_particle_coordinates(),
                field,
                wfield,
                self._hci,
                padded_ret_array,
            )

        tstop = monotonic_ns()
        if verbose:
            print(
                f"Deposited {self.particles.count:.4g} particles in {(tstop - tstart) / 1e9:.2f} s"
            )

        if weight_field is not None:
            self._apply_boundary_conditions(
                array=wfield_dep,
                boundaries=weight_field_boundaries,
                weight_array=None,
            )

            self._apply_boundary_conditions(
                array=padded_ret_array,
                boundaries=boundaries,
                weight_array=wfield_dep,
            )
        else:
            self._apply_boundary_conditions(
                array=padded_ret_array,
                boundaries=boundaries,
                weight_array=None,
            )

        with np.errstate(invalid="ignore"):
            padded_ret_array /= wfield_dep

        if return_ghost_padded_array:
            return padded_ret_array
        elif self.grid.ndim == 1:
            return padded_ret_array[1:-1]
        elif self.grid.ndim == 2:
            return padded_ret_array[1:-1, 1:-1]
        elif self.grid.ndim == 3:
            return padded_ret_array[1:-1, 1:-1, 1:-1]
        else:  # pragma: no cover
            raise RuntimeError("Caching error. Please report this.")

    def _sanitize_boundaries(self, boundaries: dict[Name, tuple[Name, Name]]) -> None:
        for ax in self.grid.axes:
            boundaries.setdefault(ax, ("open", "open"))
        for axk in boundaries:
            if axk not in self.grid.axes:
                raise ValueError(
                    f"Got invalid ax key {axk!r}, expected any of {self.grid.axes!r}"
                )
        for bound in boundaries.values():
            if (
                (not isinstance(bound, tuple))
                or len(bound) != 2
                or not all(isinstance(_, str) for _ in bound)
            ):
                raise TypeError(f"Expected a 2-tuple of strings, got {bound!r}")

            for b in bound:
                if b not in self.boundary_recipes:
                    raise ValueError(f"Unknown boundary type {b!r}")

    def _apply_boundary_conditions(
        self,
        array: NDArray[FloatT],
        boundaries: dict[Name, tuple[Name, Name]],
        weight_array: NDArray[FloatT] | None,
    ) -> None:
        axes = list(self.grid.axes)
        for ax, bv in boundaries.items():
            # in some applications, it is *crucial* that boundaries be applied
            # in a specific order, so we take advantage of the fact that dictionaries
            # are ordered in modern Python and loop over user-specified constraints
            # as a way to *expose* ordering.
            iax = axes.index(ax)
            bcs = tuple(self.boundary_recipes[key] for key in bv)
            for side, bc in zip(("left", "right"), bcs, strict=True):
                side = cast("Literal['left', 'right']", side)
                active_index: int = 1 if side == "left" else -2
                same_side_active_layer_idx = [slice(None)] * self.grid.ndim
                same_side_active_layer_idx[iax] = active_index  # type:ignore [call-overload]

                same_side_ghost_layer_idx = [slice(None)] * self.grid.ndim
                # f(-2)=-1, f(1)=0
                same_side_ghost_layer_idx[iax] = -((active_index + 1) % 2)  # type:ignore [call-overload]

                opposite_side_active_layer_idx = [slice(None)] * self.grid.ndim
                # f(-2)=1, f(1)=-2
                opposite_side_active_layer_idx[iax] = (  # type:ignore [call-overload]
                    1 if active_index == -2 else -2
                )

                opposite_side_ghost_layer_idx = [slice(None)] * self.grid.ndim
                # f(-2)=0, f(1)=-1
                opposite_side_ghost_layer_idx[iax] = -(active_index % 2)  # type:ignore [call-overload]

                if weight_array is None:
                    ONE = np.array((1,), dtype=array.dtype)
                    array[tuple(same_side_active_layer_idx)] = bc(
                        array[tuple(same_side_active_layer_idx)],
                        array[tuple(same_side_ghost_layer_idx)],
                        array[tuple(opposite_side_active_layer_idx)],
                        array[tuple(opposite_side_ghost_layer_idx)],
                        ONE,
                        ONE,
                        ONE,
                        ONE,
                        side,
                        self.metadata,
                    )
                else:
                    array[tuple(same_side_active_layer_idx)] = bc(
                        array[tuple(same_side_active_layer_idx)],
                        array[tuple(same_side_ghost_layer_idx)],
                        array[tuple(opposite_side_active_layer_idx)],
                        array[tuple(opposite_side_ghost_layer_idx)],
                        weight_array[tuple(same_side_active_layer_idx)],
                        weight_array[tuple(same_side_ghost_layer_idx)],
                        weight_array[tuple(opposite_side_active_layer_idx)],
                        weight_array[tuple(opposite_side_ghost_layer_idx)],
                        side,
                        self.metadata,
                    )



from textwrap import dedent

import numpy as np

import gpgi


def test_repr():
    ds = gpgi.load(
        geometry="cartesian",
        grid={"cell_edges": {"x": np.array([0.0, 1.0])}},
        particles={"coordinates": {"x": np.array([0.5])}},
    )

    expected_grid_repr = dedent(
        """\
        Grid(
            geometry='cartesian',
            cell_edges={'x': array([0., 1.])},
            fields={},
        )"""
    )
    assert repr(ds.grid) == expected_grid_repr

    expected_particles_repr = dedent(
        """\
        ParticleSet(
            geometry='cartesian',
            coordinates={'x': array([0.5])},
            fields={},
        )"""
    )
    assert repr(ds.particles) == expected_particles_repr

    expected_ds_repr = dedent(
        """\
        Dataset(
            geometry='cartesian',
            grid=Grid(
                geometry='cartesian',
                cell_edges={'x': array([0., 1.])},
                fields={},
            ),
            particles=ParticleSet(
                geometry='cartesian',
                coordinates={'x': array([0.5])},
                fields={},
            ),
            metadata={},
        )"""
    )
    assert repr(ds) == expected_ds_repr



import numpy as np
from numpy.typing import NDArray

NDArrayReal = NDArray[np.float64 | np.float32]
NDArrayUInt = NDArray[np.uint16]

def _index_particles(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    dx: NDArrayReal,
    out: NDArrayUInt,
) -> None: ...
def _deposit_ngp_1D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_ngp_2D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_ngp_3D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_cic_1D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_cic_2D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_cic_3D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_tsc_1D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_tsc_2D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...
def _deposit_tsc_3D(
    cell_edges_x1: NDArrayReal,
    cell_edges_x2: NDArrayReal,
    cell_edges_x3: NDArrayReal,
    particles_x1: NDArrayReal,
    particles_x2: NDArrayReal,
    particles_x3: NDArrayReal,
    field: NDArrayReal,
    weight_field: NDArrayReal,
    hci: NDArrayUInt,  # dim: 2
    out: NDArrayReal,
) -> None: ...



r"""Define public type annotations and protocols."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

# re-exports
from gpgi._typing import FieldMap

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from gpgi._typing import FloatT, HCIArray


__all__ = [
    "DepositionMethodT",
    "DepositionMethodWithMetadataT",
    "FieldMap",
]


class DepositionMethodT(Protocol):
    def __call__(  # noqa D102
        self,
        cell_edges_x1: NDArray[FloatT],
        cell_edges_x2: NDArray[FloatT],
        cell_edges_x3: NDArray[FloatT],
        particles_x1: NDArray[FloatT],
        particles_x2: NDArray[FloatT],
        particles_x3: NDArray[FloatT],
        field: NDArray[FloatT],
        weight_field: NDArray[FloatT],
        hci: HCIArray,
        out: NDArray[FloatT],
    ) -> None: ...


class DepositionMethodWithMetadataT(Protocol):
    def __call__(  # noqa D102
        self,
        cell_edges_x1: NDArray[FloatT],
        cell_edges_x2: NDArray[FloatT],
        cell_edges_x3: NDArray[FloatT],
        particles_x1: NDArray[FloatT],
        particles_x2: NDArray[FloatT],
        particles_x3: NDArray[FloatT],
        field: NDArray[FloatT],
        weight_field: NDArray[FloatT],
        hci: HCIArray,
        out: NDArray[FloatT],
        *,
        metadata: dict[str, Any],
    ) -> None: ...



import pytest

from gpgi._boundaries import BoundaryRegistry


def test_boundary_register_overrides():
    registry = BoundaryRegistry()

    def test_recipe1(
        same_side_active_layer,
        same_side_ghost_layer,
        opposite_side_active_layer,
        opposite_side_ghost_layer,
        weight_same_side_active_layer,
        weight_same_side_ghost_layer,
        weight_opposite_side_active_layer,
        weight_opposite_side_ghost_layer,
        side,
        metadata,
    ): ...
    def test_recipe2(
        same_side_active_layer,
        same_side_ghost_layer,
        opposite_side_active_layer,
        opposite_side_ghost_layer,
        weight_same_side_active_layer,
        weight_same_side_ghost_layer,
        weight_opposite_side_active_layer,
        weight_opposite_side_ghost_layer,
        side,
        metadata,
    ): ...

    registry.register("test1", test_recipe1)
    assert registry["test1"] is test_recipe1

    # registering the same function a second time shouldn't raise
    registry.register("test1", test_recipe1)

    with pytest.raises(
        ValueError,
        match="Another function is already registered with key='test1'",
    ):
        registry.register("test1", test_recipe2)

    # check that we raised in time to preserve state
    assert registry["test1"] is test_recipe1

    # if we explicitly allow unsafe mutations, this should not raise
    registry.register("test1", test_recipe2, allow_unsafe_override=True)
    assert registry["test1"] is test_recipe2



import re
from contextlib import nullcontext
from copy import deepcopy
from functools import partial
from threading import Lock

import numpy as np
import numpy.testing as npt
import pytest

import gpgi
from gpgi._lib import _deposit_ngp_2D


@pytest.fixture()
def sample_2D_dataset():
    nparticles = 100
    nx, ny = grid_shape = 16, 16
    prng = np.random.RandomState(0)

    return gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(-1, 1, nx + 1),
                "y": np.linspace(0, 2, ny + 1),
            },
            "fields": {
                "density": np.ones(grid_shape),
            },
        },
        particles={
            "coordinates": {
                "x": (2 * prng.random_sample(nparticles) - 1),
                "y": (2 * prng.random_sample(nparticles)),
            },
            "fields": {
                "mass": np.ones(nparticles),
                "vx": (2 * prng.random_sample(nparticles) - 1),
                "vy": (2 * prng.random_sample(nparticles) - 1),
            },
        },
    )


@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
def test_single_cell_grid(method):
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.array([-1.0, 1.0]),
            },
        },
        particles={
            "coordinates": {"x": 0.1 * np.arange(10, dtype="float64") - 0.5},
            "fields": {"mass": np.ones(10, dtype="float64")},
        },
    )
    out = ds.deposit("mass", method=method, return_ghost_padded_array=True)
    assert np.sum(out) == ds.particles.count


def test_missing_particles():
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {"x": np.arange(10, dtype="float64")},
        },
    )
    with pytest.raises(
        TypeError, match="Cannot deposit particle fields on a particle-less dataset"
    ):
        ds.deposit("mass", method="ngp")


def test_missing_fields():
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {"x": np.arange(10, dtype="float64")},
        },
        particles={"coordinates": {"x": np.arange(10, dtype="float64")}},
    )
    with pytest.raises(TypeError, match="There are no particle fields"):
        ds.deposit("mass", method="ngp")


def test_unknown_field(sample_2D_dataset):
    with pytest.raises(ValueError, match="Unknown particle field 'density'"):
        sample_2D_dataset.deposit("density", method="ngp")


def test_unknown_method(sample_2D_dataset):
    with pytest.raises(
        ValueError, match="Unknown deposition method 'test', expected any of (.*)"
    ):
        sample_2D_dataset.deposit("density", method="test")


def test_callable_method(sample_2D_dataset):
    sample_2D_dataset.deposit("mass", method=_deposit_ngp_2D)


def test_callable_method_with_metadata(sample_2D_dataset):
    _is_used = False
    _md_is_received = False

    def fake_dep(*args, metadata=None, **kwargs):
        nonlocal _is_used, _md_is_received
        _is_used = True
        _md_is_received = metadata is not None
        return _deposit_ngp_2D(*args, **kwargs)

    sample_2D_dataset.deposit("mass", method=fake_dep)
    assert _is_used
    assert _md_is_received


@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
@pytest.mark.mpl_image_compare
def test_2D_deposit(sample_2D_dataset, method):
    from matplotlib.figure import Figure

    ds = sample_2D_dataset
    particle_density = ds.deposit("mass", method=method)

    fig = Figure()
    ax = fig.add_subplot()

    im = ax.pcolormesh(
        "x",
        "y",
        particle_density.T,
        data=ds.grid.cell_edges,
        cmap="viridis",
        edgecolors="white",
    )
    ax.set(aspect=1, xlabel="x", ylabel="y", title=f"Deposition method '{method}'")
    ax.scatter(
        "x",
        "y",
        data=ds.particles.coordinates,
        edgecolor="black",
        color="tab:red",
        marker="o",
    )
    fig.colorbar(im, ax=ax)
    return fig


@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
@pytest.mark.parametrize("grid_type", ["linear", "geometric"])
@pytest.mark.mpl_image_compare
def test_1D_deposit(method, grid_type):
    from matplotlib.figure import Figure

    if grid_type == "linear":
        xedges = np.linspace(1, 2, 6)
    elif grid_type == "geometric":
        xedges = np.geomspace(1, 2, 6)
    else:
        raise RuntimeError  # pragma: no cover

    npart = 16
    prng = np.random.RandomState(0)
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {"x": xedges},
        },
        particles={
            "coordinates": {"x": 1 + prng.random_sample(npart)},
            "fields": {"mass": np.ones(npart)},
        },
    )
    mass = ds.deposit("mass", method=method)
    if method == "ngp":
        assert mass.sum() == ds.particles.count

    fig = Figure()
    ax = fig.add_subplot()
    ax.set(xlabel="x", ylabel="particle mass", title=f"Deposition method '{method}'")
    ax.bar(
        ds.grid.cell_centers["x"], mass, width=ds.grid.cell_widths["x"], edgecolor=None
    )
    for x in ds.particles.coordinates["x"]:
        ax.axvline(x, ls="--", color="black", lw=0.4, alpha=0.6)
    return fig


@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_3D_deposit(method, dtype):
    npart = 60
    prng = np.random.RandomState(0)
    data_2D = {
        "geometry": "cartesian",
        "grid": {
            "cell_edges": {
                "x": np.linspace(-1, 1, 10, dtype=dtype),
                "y": np.linspace(-1, 1, 10, dtype=dtype),
            },
        },
        "particles": {
            "coordinates": {
                "x": 2 * (prng.random_sample(npart).astype(dtype) - 0.5),
                "y": 2 * (prng.random_sample(npart).astype(dtype) - 0.5),
            },
            "fields": {
                "mass": np.ones(npart, dtype),
            },
        },
    }
    ds2D = gpgi.load(**data_2D)
    assert ds2D.grid.ndim == 2

    data_3D = deepcopy(data_2D)
    data_3D["grid"]["cell_edges"]["z"] = np.linspace(-1, 1, 10, dtype=dtype)
    data_3D["particles"]["coordinates"]["z"] = 2 * (
        prng.random_sample(npart).astype(dtype) - 0.5
    )
    ds3D = gpgi.load(**data_3D)
    assert ds3D.grid.ndim == 3

    # check by comparing projected 3D with direct 2D
    full_deposit_2D = ds2D.deposit(
        "mass", method=method, return_ghost_padded_array=True
    )
    full_deposit_3D = ds3D.deposit(
        "mass", method=method, return_ghost_padded_array=True
    )
    reduced_deposit = full_deposit_3D.sum(axis=2)

    rtol = 5e-7 if dtype == np.dtype("float32") else 5e-16
    npt.assert_allclose(reduced_deposit, full_deposit_2D, rtol=rtol)

    # for coverage, check that 3D deposition also works if we don't
    # return ghost layers
    deposit_3D = ds3D.deposit("mass", method=method)
    assert deposit_3D.shape == tuple(a - 2 for a in full_deposit_3D.shape)


@pytest.mark.mpl_image_compare
def test_readme_example():
    from matplotlib.figure import Figure

    nx = ny = 64
    nparticles = 600_000

    prng = np.random.RandomState(0)

    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(-1, 1, nx),
                "y": np.linspace(-1, 1, ny),
            },
        },
        particles={
            "coordinates": {
                "x": 2 * (prng.normal(0.5, 0.25, nparticles) % 1 - 0.5),
                "y": 2 * (prng.normal(0.5, 0.25, nparticles) % 1 - 0.5),
            },
            "fields": {
                "mass": np.ones(nparticles),
            },
        },
    )

    particle_mass = ds.deposit("mass", method="nearest_grid_point")

    fig = Figure()
    ax = fig.add_subplot()
    ax.set(aspect=1, xlabel="x", ylabel="y")

    im = ax.pcolormesh(
        "x",
        "y",
        particle_mass.T,
        data=ds.grid.cell_edges,
        cmap="viridis",
    )
    fig.colorbar(im, ax=ax, label="deposited particle mass")
    return fig


def test_performance_logging(capsys, sample_2D_dataset):
    ds = sample_2D_dataset
    ds.deposit("mass", method="ngp", verbose=True)
    stdout, stderr = capsys.readouterr()
    assert stderr == ""
    lines = stdout.strip().split("\n")
    assert len(lines) == 2
    assert re.match(r"Indexed .* particles in .* s", lines[0])
    assert re.match(r"Deposited .* particles in .* s", lines[1])


def test_return_ghost_padded_array():
    npart = 16
    prng = np.random.RandomState(0)
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {"x": np.linspace(1, 2, 6)},
        },
        particles={
            "coordinates": {"x": 1 + prng.random_sample(npart)},
            "fields": {"mass": np.ones(npart)},
        },
    )
    active_array = ds.deposit("mass", method="cic")
    assert active_array.shape == (5,)
    padded_array = ds.deposit(
        "mass", method="cic", verbose=True, return_ghost_padded_array=True
    )
    assert padded_array.shape == (7,)


@pytest.mark.parametrize(
    "boundaries, error_type, error_message",
    [
        (
            {"radius": ("open", "open")},
            ValueError,
            "Got invalid ax key 'radius', expected any of ('x', 'y')",
        ),
        (
            {"x": ("o", "o", "o")},
            TypeError,
            "Expected a 2-tuple of strings, got ('o', 'o', 'o')",
        ),
        (
            {"x": ("open", "unregistered_key")},
            ValueError,
            "Unknown boundary type 'unregistered_key'",
        ),
    ],
)
def test_deposit_invalid_boundaries(
    boundaries, error_type, error_message, sample_2D_dataset
):
    ds = sample_2D_dataset
    with pytest.raises(error_type, match=re.escape(error_message)):
        ds.deposit("mass", method="ngp", boundaries=boundaries)


@pytest.mark.parametrize(
    "keyL, keyR",
    [("open", "wall"), ("periodic", "periodic"), ("antisymmetric ", "open")],
)
def test_builtin_boundary_recipe(keyL, keyR, sample_2D_dataset):
    ds = sample_2D_dataset
    res1 = ds.deposit("mass", method="tsc", boundaries={"x": (keyL, keyR)})

    if keyR == keyL:
        return
    res2 = ds.deposit("mass", method="tsc", boundaries={"x": (keyR, keyL)})
    npt.assert_array_equal(res1[1:-1, 1:-1], res2[1:-1, 1:-1])


def test_register_invalid_boundary_recipe():
    nx = ny = 64
    nparticles = 100

    prng = np.random.RandomState(0)
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(-1, 1, nx),
                "y": np.linspace(-1, 1, ny),
            },
        },
        particles={
            "coordinates": {
                "x": 2 * (prng.normal(0.5, 0.25, nparticles) % 1 - 0.5),
                "y": 2 * (prng.normal(0.5, 0.25, nparticles) % 1 - 0.5),
            },
            "fields": {
                "mass": np.ones(nparticles),
            },
        },
    )

    def _my_recipe(a, b, c, d, e, f):
        return a  # pragma: no cover

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Invalid boundary recipe. Expected a function with exactly 6 parameters, "
            "named 'same_side_active_layer', 'same_side_ghost_layer', "
            "'opposite_side_active_layer', 'opposite_side_ghost_layer', "
            "'weight_same_side_active_layer', 'weight_same_side_ghost_layer', "
            "'weight_opposite_side_active_layer', 'weight_opposite_side_ghost_layer', "
            "'side', and 'metadata'"
        ),
    ):
        ds.boundary_recipes.register("my", _my_recipe)


def test_register_custom_boundary_recipe(sample_2D_dataset):
    def _my_recipe(
        same_side_active_layer,
        same_side_ghost_layer,
        opposite_side_active_layer,
        opposite_side_ghost_layer,
        weight_same_side_active_layer,
        weight_same_side_ghost_layer,
        weight_opposite_side_active_layer,
        weight_opposite_side_ghost_layer,
        side,
        metadata,
    ):
        # return the active layer unchanged
        # (this is the same as the builtin 'open' boundary recipe)
        return same_side_active_layer

    ds = sample_2D_dataset
    ds.boundary_recipes.register("my", _my_recipe)
    res1 = ds.deposit("mass", method="tsc", boundaries={"x": ("my", "my")})
    res2 = ds.deposit("mass", method="tsc")
    npt.assert_array_equal(res1, res2)


@pytest.fixture
def unary_mass_dataset():
    from itertools import product

    x = np.linspace(0.5, 9.5, 10)
    xc, yc, zc = np.array(list(product(x, x, x))).T

    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(0, 10, 11),
                "y": np.linspace(0, 10, 11),
                "z": np.linspace(0, 10, 11),
            },
        },
        particles={
            "coordinates": {
                "x": xc,
                "y": yc,
                "z": zc,
            },
            "fields": {"mass": 1e-3 * np.ones(1000)},
        },
    )
    return ds


@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
def test_unary_mass_dataset(method, unary_mass_dataset):
    # check that total mass is conserved with open boundaries
    # when including ghost zones
    ds = unary_mass_dataset
    base_array = ds.deposit("mass", method=method, return_ghost_padded_array=True)
    expected = ds.particles.fields["mass"].sum()
    assert base_array.sum() == pytest.approx(expected, rel=1e-12)


@pytest.mark.parametrize(
    "boundaries",
    [
        {
            "x": ("periodic", "periodic"),
            "y": ("periodic", "periodic"),
            "z": ("periodic", "periodic"),
        },
        {
            "x": ("wall", "wall"),
            "y": ("wall", "wall"),
            "z": ("wall", "wall"),
        },
    ],
)
@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
def test_closed_boundaries(method, boundaries, unary_mass_dataset):
    ds = unary_mass_dataset
    arr = ds.deposit("mass", method=method, boundaries=boundaries)
    expected = ds.particles.fields["mass"].sum()
    assert arr.sum() == pytest.approx(expected, rel=1e-12)


@pytest.mark.skipif(
    gpgi._IS_PY_LIB,
    reason="weight_field is not implemented with gpgi._IS_PY_LIB = True",
)
def test_deposit_with_weight_field(sample_2D_dataset):
    sample_2D_dataset.deposit("vx", method="tsc", weight_field="mass")


def test_deposit_with_weight_field_and_incomplete_boundaries_spec(sample_2D_dataset):
    with pytest.raises(
        TypeError,
        match="weight_field_boundaries keyword argument is required with weight_field and boundaries",
    ):
        sample_2D_dataset.deposit(
            "vx",
            method="tsc",
            weight_field="mass",
            boundaries={"x": ("periodic", "periodic")},
        )


def test_deposit_with_weight_field_and_complete_boundaries_spec(sample_2D_dataset):
    ds = sample_2D_dataset
    ds.deposit(
        "vx",
        method="ngp",
        weight_field="mass",
        boundaries={"x": ("periodic", "periodic")},
        weight_field_boundaries={"x": ("periodic", "periodic")},
    )


def test_warn_unused_weight_field_boundaries(sample_2D_dataset):
    with pytest.warns(
        UserWarning,
        match=(
            "weight_field_boundaries will not be used as no weight_field was specified"
        ),
    ):
        sample_2D_dataset.deposit(
            "vx",
            method="ngp",
            weight_field_boundaries={"x": ("periodic", "periodic")},
        )


def test_partial_boundary():
    def _base_recipe(
        same_side_active_layer,
        same_side_ghost_layer,
        opposite_side_active_layer,
        opposite_side_ghost_layer,
        weight_same_side_active_layer,
        weight_same_side_ghost_layer,
        weight_opposite_side_active_layer,
        weight_opposite_side_ghost_layer,
        side,
        metadata,
        *,
        mode,  # supplementary argument that is destined to be frozen with functools.partial
    ):
        # the return value doesn't matter, this test checks that
        # recipe validation doesn't fail
        return same_side_active_layer  # pragma: no cover

    myrecipe = partial(_base_recipe, mode="test")

    # define a minimal single-use dataset
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(0, 10, 11),
                "y": np.linspace(0, 10, 11),
                "z": np.linspace(0, 10, 11),
            },
        },
    )
    ds.boundary_recipes.register("my", myrecipe, skip_validation=False)


@pytest.mark.parametrize("method", ["ngp", "cic", "tsc"])
@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_particles_on_domain_corners(method, dtype):
    nx, ny, nz = 2, 3, 5
    xlim = ylim = zlim = (0, 1)
    nparticles = 6
    ds = gpgi.load(
        geometry="cartesian",
        grid={
            "cell_edges": {
                "x": np.linspace(*xlim, nx + 1, dtype=dtype),
                "y": np.linspace(*ylim, ny + 1, dtype=dtype),
                "z": np.linspace(*zlim, nz + 1, dtype=dtype),
            },
        },
        particles={
            "coordinates": {
                "x": np.array([0, 0, 0, 1, 1, 1], dtype=dtype),
                "y": np.array([0, 1, 1, 0, 1, 1], dtype=dtype),
                "z": np.array([1, 0, 1, 0, 0, 1], dtype=dtype),
            },
            "fields": {
                "mass": np.ones(nparticles, dtype=dtype),
            },
        },
    )
    ds.deposit("mass", method=method)


@pytest.mark.parametrize("lock", ["per-instance", None, Lock()])
def test_explicit_lock(lock, sample_2D_dataset):
    ds = sample_2D_dataset
    mass_dep_ref = ds.deposit("mass", method="nearest_grid_point")
    mass_dep_exp = ds.deposit("mass", method="nearest_grid_point", lock=lock)
    npt.assert_array_equal(mass_dep_exp, mass_dep_ref)


@pytest.mark.parametrize("lock", ["perinstance", 0, nullcontext()])
def test_invalick_lock(lock, sample_2D_dataset):
    ds = sample_2D_dataset
    with pytest.raises(ValueError, match="Received lock="):
        ds.deposit("mass", method="nearest_grid_point", lock=lock)



import re

import numpy as np
import pytest

import gpgi


def get_random_dataset(dimensionality: int):
    NPARTICLES = 10_000
    NX = 16
    prng = np.random.RandomState(0)

    cell_edges = {
        "x": np.linspace(-1, 1, NX + 1),
    }
    coordinates = {"x": (2 * prng.random_sample(NPARTICLES) - 1)}
    if dimensionality >= 2:
        cell_edges["y"] = np.linspace(-1, 1, NX + 1)
        coordinates["y"] = 2 * prng.random_sample(NPARTICLES) - 1
    if dimensionality == 3:
        cell_edges["z"] = np.linspace(-1, 1, NX + 1)
        coordinates["z"] = 2 * prng.random_sample(NPARTICLES) - 1

    return gpgi.load(
        geometry="cartesian",
        grid={"cell_edges": cell_edges},
        particles={"coordinates": coordinates},
    )


@pytest.mark.parametrize("dim", [1, 2, 3])
def test_sort(dim):
    ds = get_random_dataset(dim)
    assert not ds.is_sorted()
    sds = ds.sorted()
    assert sds.is_sorted()


@pytest.mark.parametrize(
    "axes, expected_message",
    [
        ((0, 1), "Expected exactly 3 axes, got 2"),
        (
            (0.0, 1.0, 2.0),
            (
                "Expected all axes to be integers, got (0.0, 1.0, 2.0) "
                "with types (float, float, float)"
            ),
        ),
        ((1, 2, 3), "Expected all axes to be <3, got (1, 2, 3)"),
    ],
)
def test_key_errors(axes, expected_message):
    ds = get_random_dataset(3)
    with pytest.raises(ValueError, match=re.escape(expected_message)):
        ds.sorted(axes=axes)



"""gpgi: Fast particle deposition at post-processing time."""

from importlib.util import find_spec

from ._data_types import Dataset, Geometry, Grid, ParticleSet
from ._load import load

__all__ = [
    "load",
    "Dataset",
    "Geometry",
    "Grid",
    "ParticleSet",
]

_IS_PY_LIB = find_spec("gpgi._lib").origin.endswith(".py")  # type: ignore [union-attr] # pyright: ignore [reportOptionalMemberAccess]
